# -*- coding: utf-8 -*-
"""
.. module:: MyCapytain.retrievers.dts
   :synopsis: DTS endpoint implementation

.. moduleauthor:: Thibault Clérice <leponteineptique@gmail.com>


"""
import MyCapytain.retrievers.prototypes
from MyCapytain import __version__
from MyCapytain.common.reference import BaseReference
import requests
from MyCapytain.common.utils import parse_uri


__all__ = [
    "HttpDtsRetriever"
]


def _parse_ref_parameters(parameters, ref):
    if isinstance(ref, BaseReference):
        if ref.is_range():
            parameters["start"], parameters["end"] = ref
        else:
            parameters["ref"] = ref.start
    elif ref:
        parameters["ref"] = ref


class HttpDtsRetriever(MyCapytain.retrievers.prototypes.API):
    def __init__(self, endpoint):
        super(HttpDtsRetriever, self).__init__(endpoint)
        self._routes = None

    def call(self, route, parameters, mimetype="application/ld+json", defaults=None):
        """ Call an endpoint given the parameters

        :param route: Named of the route which is called
        :type route: str
        :param parameters: Dictionary of parameters
        :type parameters: dict
        :param mimetype: Mimetype to require
        :type mimetype: str
        :rtype: text
        """
        if not defaults:
            defaults = {}
        parameters = {
            key: str(parameters[key])
            for key in parameters
            if parameters[key] is not None and
               parameters[key] != defaults.get(key, None)
        }
        parameters.update(self.routes[route].query_dict)

        request = requests.get(
            self.routes[route].path,
            params=parameters,
            headers={
                "Accept": mimetype,
                "Accept-Charset": "utf-8",
                "User-Agent": "MyCapytain/{MyCapVersion} {DefaultRequestUA}".format(
                    MyCapVersion=__version__,
                    DefaultRequestUA=requests.utils.default_user_agent()
                )
            }
        )
        request.raise_for_status()
        if request.encoding is None:
            request.encoding = "utf-8"
        return request

    @property
    def routes(self):
        """ Retrieves the main routes of the DTS Collection

        Response format expected :
            {
              "@context": "/dts/api/contexts/EntryPoint.jsonld",
              "@id": "/dts/api/",
              "@type": "EntryPoint",
              "collections": "/dts/api/collections/",
              "documents": "/dts/api/documents/",
              "navigation" : "/dts/api/navigation"
            }

        :returns: Dictionary of main routes with their path
        :rtype: dict
        """
        if self._routes:
            return self._routes

        request = requests.get(self.endpoint)
        request.raise_for_status()
        data = request.json()
        self._routes = {
            "collections": parse_uri(data["collections"], self.endpoint),
            "documents": parse_uri(data["documents"], self.endpoint),
            "navigation": parse_uri(data["navigation"], self.endpoint)
        }
        return self._routes

    def get_collection(self, collection_id=None, nav="children", page=None):
        """ Makes a call on the Collection API

        :param collection_id: Id of the collection to retrieve
        :param nav: Direction of the navigation
        :param page: Page to retrieve
        :return: Response
        :rtype: requests.Response
        """
        return self.call(
            "collections",
            {
                "id": collection_id,
                "nav": nav,
                "page": page
            },
            defaults={
                "id": None,
                "nav": "children",
                "page": 1
            }
        )

    def get_navigation(
            self, collection_id, level=None, ref=None,
            group_by=None, max_=None, exclude=None, page=None):
        """ Make a navigation request on the DTS API

        :param collection_id: Id of the collection
        :param level: Lever at which the references should be listed
        :param ref: If ref is a tuple, it is treated as a range. String or int are treated as single ref
        :param group_by: Size of the ranges the server should produce
        :param max_: Maximum number of results
        :param exclude: Exclude specific metadata.
        :param page: Page
        :return: Response
        :rtype: requests.Response
        """
        parameters = {
            "id": collection_id,
            "level": level,
            "groupBy": group_by,
            "max": max_,
            "exclude": exclude,
            "page": page
        }
        _parse_ref_parameters(parameters, ref)

        return self.call(
            "navigation",
            parameters
        )

    def get_document(
            self,
            collection_id, ref=None, mimetype="application/tei+xml, application/xml"):
        """ Make a document request on the DTS API

        :param collection_id: Id of the collection
        :param ref: If ref is a tuple, it is treated as a range. String or int are treated as single ref
        :param mimetype: Media type to request
        :return: Response
        :rtype: requests.Response
        """
        parameters = {
            "id": collection_id
        }
        _parse_ref_parameters(parameters, ref)

        return self.call(
            "documents",
            parameters,
            mimetype=mimetype
        )



# -*- coding: utf-8 -*-
"""
.. module:: MyCapytain.resolvers.cts.remote
   :synopsis: Resolver built for CTS APIs

.. moduleauthor:: Thibault Clérice <leponteineptique@gmail.com>


"""

from MyCapytain.resolvers.prototypes import Resolver
from MyCapytain.resources.texts.remote.cts import CtsText
from MyCapytain.resources.collections.cts import XmlCtsTextInventoryMetadata
from MyCapytain.retrievers.cts5 import HttpCtsRetriever


__all__ = [
    "HttpCtsResolver"
]

class HttpCtsResolver(Resolver):
    """ HttpCtsResolver provide a resolver for CTS API http endpoint.

    :param endpoint: CTS API Retriever
    :type endpoint: CTS

    :ivar endpoint: CTS API Retriever
    :type endpoint: HttpCtsRetriever
    """
    def __init__(self, endpoint):
        if not isinstance(endpoint, HttpCtsRetriever):
            raise TypeError("Endpoint should be a CTS Endpoint object")
        self.__endpoint__ = endpoint

    @property
    def endpoint(self):
        """ CTS Endpoint of the resolver

        :return: CTS Endpoint
        :rtype: HttpCtsRetriever
        """
        return self.__endpoint__

    def getTextualNode(self, textId, subreference=None, prevnext=False, metadata=False):
        """ Retrieve a text node from the API

        :param textId: CtsTextMetadata Identifier
        :type textId: str
        :param subreference: CapitainsCtsPassage Reference
        :type subreference: str
        :param prevnext: Retrieve graph representing previous and next passage
        :type prevnext: boolean
        :param metadata: Retrieve metadata about the passage and the text
        :type metadata: boolean
        :return: CapitainsCtsPassage
        :rtype: CapitainsCtsPassage
        """
        text = CtsText(
            urn=textId,
            retriever=self.endpoint
        )
        if metadata or prevnext:
            return text.getPassagePlus(reference=subreference)
        else:
            return text.getTextualNode(subreference=subreference)

    def getSiblings(self, textId, subreference):
        """ Retrieve the siblings of a textual node

        :param textId: CtsTextMetadata Identifier
        :type textId: str
        :param subreference: CapitainsCtsPassage Reference
        :type subreference: str
        :return: Tuple of references
        :rtype: (str, str)
        """
        text = CtsText(
            urn=textId,
            retriever=self.endpoint
        )
        return text.getPrevNextUrn(subreference)

    def getReffs(self, textId, level=1, subreference=None):
        """ Retrieve the siblings of a textual node

        :param textId: CtsTextMetadata Identifier
        :type textId: str
        :param level: Depth for retrieval
        :type level: int
        :param subreference: CapitainsCtsPassage Reference
        :type subreference: str
        :return: List of references
        :rtype: [str]
        """
        text = CtsText(
            urn=textId,
            retriever=self.endpoint
        )
        return text.getReffs(level, subreference)

    def getMetadata(self, objectId=None, **filters):
        """ Request metadata about a text or a collection

        :param objectId: Object Identifier to filter on
        :type objectId: str
        :param filters: Kwargs parameters.
        :type filters: dict
        :return: Collection
        """
        if objectId is not None:
            filters["urn"] = objectId

        ti = XmlCtsTextInventoryMetadata.parse(self.endpoint.getCapabilities(**filters))
        if objectId:
            return [x for x in [ti] + ti.descendants if x.id == objectId][0]
        return ti



# -*- coding: utf-8 -*-
"""
.. module:: MyCapytain.common.utils
   :synopsis: Common useful tools

.. moduleauthor:: Thibault Clérice <leponteineptique@gmail.com>


"""

from ._generic import (
    OrderedDefaultDict, nested_ordered_dictionary, nested_get, nested_set, normalize
)
from ._graph import Subgraph, expand_namespace
from ._http import parse_pagination, parse_uri
from ._json_ld import dict_to_literal, literal_to_dict



from collections import defaultdict

from rdflib import Graph, BNode
from rdflib.namespace import NamespaceManager


class Subgraph(object):
    """ Utility class to generate subgraph around one or more items

    :param
    """
    def __init__(self, bindings: dict = None):
        self.graph = Graph()
        self.graph.namespace_manager = NamespaceManager(self.graph)
        for prefix, namespace in bindings.items():
            self.graph.namespace_manager.bind(prefix, namespace)

        self.downwards = defaultdict(lambda: True)
        self.updwards = defaultdict(lambda: True)

    def graphiter(self, graph, target, ascendants=0, descendants=1):
        """ Iter on a graph to finds object connected

        :param graph: Graph to serialize
        :type graph: Graph
        :param target: Node to iterate over
        :type target: Node
        :param ascendants: Number of level to iter over upwards (-1 = No Limit)
        :param descendants: Number of level to iter over downwards (-1 = No limit)
        :return:
        """

        asc = 0 + ascendants
        if asc != 0:
            asc -= 1

        desc = 0 + descendants
        if desc != 0:
            desc -= 1

        t = str(target)

        if descendants != 0 and self.downwards[t] is True:
            self.downwards[t] = False
            for pred, obj in graph.predicate_objects(target):
                if desc == 0 and isinstance(obj, BNode):
                    continue
                self.add((target, pred, obj))

                # Retrieve triples about the object
                if desc != 0 and self.downwards[str(obj)] is True:
                    self.graphiter(graph, target=obj, ascendants=0, descendants=desc)

        if ascendants != 0 and self.updwards[t] is True:
            self.updwards[t] = False
            for s, p in graph.subject_predicates(object=target):
                if desc == 0 and isinstance(s, BNode):
                    continue
                self.add((s, p, target))

                # Retrieve triples about the parent as object
                if asc != 0 and self.updwards[str(s)] is True:
                    self.graphiter(graph, target=s, ascendants=asc, descendants=0)

    def serialize(self, *args, **kwargs):
        return self.graph.serialize(*args, **kwargs)

    def add(self, *args, **kwargs):
        self.graph.add(*args, **kwargs)


def expand_namespace(nsmap, string):
    """ If the string starts with a known prefix in nsmap, replace it by full URI

    :param nsmap: Dictionary of prefix -> uri of namespace
    :param string: String in which to replace the namespace
    :return: Expanded string with no namespace
    """
    for ns in nsmap:
        if isinstance(string, str) and isinstance(ns, str) and string.startswith(ns+":"):
            return string.replace(ns+":", nsmap[ns])
    return string


# -*- coding: utf-8 -*-
"""
.. module:: MyCapytain.retrievers
   :synopsis: Implementation and abstraction of CTS related retrievers

.. moduleauthor:: Thibault Clérice <leponteineptique@gmail.com>


"""


"""
.. module:: MyCapytain.resolvers.prototypes
   :synopsis: Resolver Prototype

.. moduleauthor:: Thibault Clérice <leponteineptique@gmail.com>


"""

from typing import Tuple, Union, Optional, Dict, Any
from MyCapytain.resources.prototypes.metadata import Collection
from MyCapytain.resources.prototypes.text import TextualNode
from MyCapytain.common.reference import BaseReference, BaseReferenceSet


__all__ = [
    "Resolver"
]


class Resolver(object):
    """ Resolver provide a native python API which returns python objects.

    Initiation of resolvers are dependent on the implementation of the prototype

    """
    def getMetadata(self, objectId: str=None, **filters) -> Collection:
        """ Request metadata about a text or a collection

        :param objectId: Object Identifier to filter on
        :type objectId: str
        :param filters: Kwargs parameters.
        :type filters: dict
        :return: Collection
        """
        raise NotImplementedError()

    def getTextualNode(
            self,
            textId: str,
            subreference: Union[str, BaseReference]=None,
            prevnext: bool=False,
            metadata: bool=False
    ) -> TextualNode:
        """ Retrieve a text node from the API

        :param textId: CtsTextMetadata Identifier
        :type textId: str
        :param subreference: CapitainsCtsPassage Reference
        :type subreference: str
        :param prevnext: Retrieve graph representing previous and next passage
        :type prevnext: boolean
        :param metadata: Retrieve metadata about the passage and the text
        :type metadata: boolean
        :return: CapitainsCtsPassage
        :rtype: CapitainsCtsPassage
        """
        raise NotImplementedError()

    def getSiblings(self, textId: str, subreference: Union[str, BaseReference]) -> Tuple[BaseReference, BaseReference]:
        """ Retrieve the siblings of a textual node

        :param textId: CtsTextMetadata Identifier
        :type textId: str
        :param subreference: CapitainsCtsPassage Reference
        :type subreference: str
        :return: Tuple of references
        :rtype: (str, str)
        """
        raise NotImplementedError()

    def getReffs(
            self,
            textId: str,
            level: int=1,
            subreference: Union[str, BaseReference]=None,
            include_descendants: bool=False,
            additional_parameters: Optional[Dict[str, Any]]=None
    ) -> BaseReferenceSet:
        """ Retrieve the siblings of a textual node

        :param textId: CtsTextMetadata Identifier
        :type textId: str
        :param level: Depth for retrieval
        :type level: int
        :param subreference: CapitainsCtsPassage Reference
        :type subreference: str
        :param include_descendants:
        :param additional_parameters:
        :return: List of references
        :rtype: [str]

        ..toDo :: This starts to be a bloated function....
        """
        raise NotImplementedError()



from MyCapytain.resources.prototypes.metadata import Collection
from MyCapytain.errors import JsonLdCollectionMissing
from MyCapytain.common.reference import DtsCitationSet
from MyCapytain.common.constants import RDF_NAMESPACES
from MyCapytain.common.utils.dts import parse_metadata


from typing import List
from pyld import jsonld


__all__ = [
    "DtsCollection"
]


_hyd = RDF_NAMESPACES.HYDRA
_dts = RDF_NAMESPACES.DTS
_cap = RDF_NAMESPACES.CAPITAINS
_tei = RDF_NAMESPACES.TEI
_empty_extensions = [{}]


class DtsCollection(Collection):

    CitationSet = DtsCitationSet

    def __init__(self, identifier="", *args, **kwargs):
        super(DtsCollection, self).__init__(identifier, *args, **kwargs)
        self._expanded = False  # Not sure I'll keep this
        self._citation = DtsCitationSet()
        self._parents = set()

    @property
    def citation(self):
        return self._citation

    @citation.setter
    def citation(self, citation: CitationSet):
        self._citation = citation

    @property
    def size(self):
        value = self.metadata.get_single(RDF_NAMESPACES.HYDRA.totalItems)
        if value:
            return int(value)
        return 0

    @property
    def readable(self):
        if self.type == RDF_NAMESPACES.HYDRA.Resource:
            return True
        return False

    @property
    def parents(self):
        return self._parents

    @parents.setter
    def parents(self, value):
        self._parents = value

    @property
    def parent(self):
        raise NotImplementedError("In DTS, parent only exists in plural, via .parents")

    @parent.setter
    def parent(self, value):
        raise NotImplementedError("In DTS, parent only exists in plural, via .parents")

    @classmethod
    def parse(cls, resource, direction="children", **additional_parameters) -> "DtsCollection":
        """ Given a dict representation of a json object, generate a DTS Collection

        :param resource:
        :type resource: dict
        :param direction: Direction of the hydra:members value
        :return: DTSCollection parsed
        :rtype: DtsCollection
        """

        data = jsonld.expand(resource)
        if len(data) == 0:
            raise JsonLdCollectionMissing("Missing collection in JSON")
        data = data[0]

        obj = cls(
            identifier=resource["@id"],
            **additional_parameters
        )

        obj._parse_metadata(data)
        obj._parse_members(data, direction=direction, **additional_parameters)

        return obj

    def _parse_members(self, data, direction: str="children", **additional_parameters: dict):
        """

        :param data:
        :param direction:
        :param additional_parameters:
        :return:
        """
        members = self.parse_member(
            data, self, direction, **additional_parameters
        )
        if direction == "children":  # ToDo: Should be in a third function ?
            self.children.update({
                coll.id: coll
                for coll in members
            })
        else:
            self.parents.add(members)

    def _parse_metadata(self, data: dict):

        # We retrieve first the descriptiooon and label that are dependant on Hydra
        for val_dict in data[str(_hyd.title)]:
            self.set_label(val_dict["@value"], None)
        for val_dict in data["@type"]:
            self.type = val_dict

        # We retrieve the Citation System
        _cite_structure_term = str(_dts.term("citeStructure"))
        if _cite_structure_term in data and data[_cite_structure_term]:
            self.citation = self.CitationSet.ingest(data[_cite_structure_term])

        _cite_depth_term = str(_dts.term("citeDepth"))
        if _cite_depth_term in data and data[_cite_depth_term]:
            self.citation.depth = data[_cite_depth_term][0]["@value"]

        for val_dict in data[str(_hyd.totalItems)]:
            self.metadata.add(_hyd.totalItems, val_dict["@value"])

        for val_dict in data.get(str(_hyd.description), []):
            self.metadata.add(_hyd.description, val_dict["@value"], None)

        parse_metadata(self.metadata, data)

    def retrieve(self) -> bool:
        """ If needed, retrieves complete metadata

        :return: Status of retrieval
        """
        return True

    @classmethod
    def parse_member(
            cls,
            obj: dict,
            collection: "DtsCollection",
            direction: str,
            **additional_parameters) -> List["DtsCollection"]:
        """ Parse the member value of a Collection response
        and returns the list of object while setting the graph
        relationship based on `direction`

        :param obj: PyLD parsed JSON+LD
        :param collection: Collection attached to the member property
        :param direction: Direction of the member (children, parent)
        """
        members = []

        for member in obj.get(str(_hyd.member), []):
            subcollection = cls.parse(member, **additional_parameters)
            if direction == "children":
                subcollection.parents.update({collection})
            members.append(subcollection)

        return members



# -*- coding: utf-8 -*-
"""
.. module:: MyCapytain.retrievers.cts5
   :synopsis: Cts5 endpoint implementation

.. moduleauthor:: Thibault Clérice <leponteineptique@gmail.com>


"""
import MyCapytain.retrievers.prototypes
from MyCapytain.common.reference._capitains_cts import CtsReference
import requests


__all__ = [
    "HttpCtsRetriever"
]


class HttpCtsRetriever(MyCapytain.retrievers.prototypes.CtsRetriever):
    """ Basic integration of the MyCapytain.retrievers.proto.CTS abstraction
    """

    def __init__(self, endpoint, inventory=None):
        """ API Prototype object

        :param self: Object
        :type self: API
        :param endpoint: URL of the API
        :type endpoint: str
        :param inventory: Inventory to use
        :type inventory: str

        :ivar endpoint: Url of the endpoint
        :ivar inventory: Default Inventory
        """
        super(HttpCtsRetriever, self).__init__(endpoint)
        self.inventory = inventory

    def call(self, parameters):
        """ Call an endpoint given the parameters

        :param parameters: Dictionary of parameters
        :type parameters: dict
        :rtype: text
        """
        # DEV !
        parameters = {
            key: str(parameters[key]) for key in parameters if parameters[key] is not None
        }
        if self.inventory is not None and "inv" not in parameters:
            parameters["inv"] = self.inventory

        request = requests.get(self.endpoint, params=parameters)
        request.raise_for_status()
        if request.encoding is None:
            request.encoding = "utf-8"
        return request.text

    def getCapabilities(self, inventory=None, urn=None):
        """ Retrieve the inventory information of an API

        :param inventory: Name of the inventory
        :type inventory: text
        :param urn: URN to filter with
        :type urn: str
        :rtype: str
        """
        return self.call({
            "inv": inventory,
            "urn": urn,
            "request": "GetCapabilities"
        })

    def getValidReff(self, urn, inventory=None, level=None):
        """ Retrieve valid urn-references for a text

        :param urn: URN identifying the text
        :type urn: text
        :param inventory: Name of the inventory
        :type inventory: text
        :param level: Depth of references expected
        :type level: int
        :return: XML Response from the API as string
        :rtype: str
        """
        return self.call({
            "inv": inventory,
            "urn": urn,
            "level": level,
            "request": "GetValidReff"
        })

    def getFirstUrn(self, urn, inventory=None):
        """ Retrieve the first passage urn of a text

        :param urn: URN identifying the text
        :type urn: text
        :param inventory: Name of the inventory
        :type inventory: text
        :rtype: str
        """
        return self.call({
            "inv": inventory,
            "urn": urn,
            "request": "GetFirstUrn"
        })

    def getPrevNextUrn(self, urn, inventory=None):
        """ Retrieve the previous and next passage urn of one passage

        :param urn: URN identifying the text's passage (Minimum depth : 1)
        :type urn: text
        :param inventory: Name of the inventory
        :type inventory: text
        :rtype: str
        """
        return self.call({
            "inv": inventory,
            "urn": urn,
            "request": "GetPrevNextUrn"
        })

    def getLabel(self, urn, inventory=None):
        """ Retrieve informations about a CTS Urn

        :param urn: URN identifying the text's passage (Minimum depth : 1)
        :type urn: text
        :param inventory: Name of the inventory
        :type inventory: text
        :rtype: str
        """
        return self.call({
            "inv": inventory,
            "urn": urn,
            "request": "GetLabel"
        })

    def getPassage(self, urn, inventory=None, context=None):
        """ Retrieve a passage

        :param urn: URN identifying the text's passage (Minimum depth : 1)
        :type urn: text
        :param inventory: Name of the inventory
        :type inventory: text
        :param context: Number of citation units at the same level of the citation hierarchy as the requested urn, immediately preceding and immediately following the requested urn to include in the reply
        :type context: int
        :rtype: str
        """
        return self.call({
            "inv": inventory,
            "urn": urn,
            "context": context,
            "request": "GetPassage"
        })

    def getPassagePlus(self, urn, inventory=None, context=None):
        """ Retrieve a passage and information about it

        :param urn: URN identifying the text's passage (Minimum depth : 1)
        :type urn: text
        :param inventory: Name of the inventory
        :type inventory: text
        :param context: Number of citation units at the same level of the citation hierarchy as the requested urn, immediately preceding and immediately following the requested urn to include in the reply
        :type context: int
        :rtype: str
        """
        return self.call({
            "inv": inventory,
            "urn": urn,
            "context": context,
            "request": "GetPassagePlus"
        })

    #
    # Common methods
    #

    def getMetadata(self, objectId=None, **filters):
        """ Request metadata about a text or a collection

        :param objectId: Filter for some object identifier
        :param filters: Kwargs parameters. URN and Inv are available
        :return: GetCapabilities CTS API request response
        """
        filters.update({"urn": objectId})
        return self.getCapabilities(**filters)

    def getTextualNode(self, textId, subreference=None, prevnext=False, metadata=False):
        """ Retrieve a text node from the API

        :param textId: CtsTextMetadata Identifier
        :param subreference: CapitainsCtsPassage Reference
        :param prevnext: Retrieve graph representing previous and next passage
        :param metadata: Retrieve metadata about the passage and the text
        :return: GetPassage or GetPassagePlus CTS API request response
        """
        if  subreference:
            textId = "{}:{}".format(textId, subreference)

        if prevnext or metadata:
            return self.getPassagePlus(urn=textId)
        else:
            return self.getPassage(urn=textId)

    def getSiblings(self, textId, subreference):
        """ Retrieve the siblings of a textual node

        :param textId: CtsTextMetadata Identifier
        :param reference: CapitainsCtsPassage Reference
        :return: GetPrevNextUrn request response from the endpoint
        """
        textId = "{}:{}".format(textId, subreference)
        return self.getPrevNextUrn(urn=textId)

    def getReffs(self, textId, level=1, subreference=None):
        """ Retrieve the siblings of a textual node

        :param textId: CtsTextMetadata Identifier
        :type textId: str
        :param level: Depth for retrieval
        :type level: int
        :param subreference: CapitainsCtsPassage Reference
        :type subreference: str
        :return: List of references
        :rtype: [str]
        """
        depth = level
        if subreference:
            textId = "{}:{}".format(textId, subreference)
        if subreference:
            if isinstance(subreference, CtsReference):
                depth += subreference.depth
            else:
                depth += (CtsReference(subreference)).depth
        if level:
            level = max(depth, level)
        return self.getValidReff(urn=textId, level=level)



# -*- coding: utf-8 -*-
"""
.. module:: MyCapytain.resources.prototypes.metadata
   :synopsis: Definition of Metadata Type Objects

.. moduleauthor:: Thibault Clérice <leponteineptique@gmail.com>

"""

from MyCapytain.common.metadata import Metadata
from MyCapytain.errors import UnknownCollection
from MyCapytain.common.utils import literal_to_dict, Subgraph
from MyCapytain.common.constants import RDF_NAMESPACES, RDFLIB_MAPPING, Mimetypes, get_graph
from MyCapytain.common.base import Exportable
from MyCapytain.common.reference import BaseCitationSet
from rdflib import URIRef, RDF, Literal, Graph, RDFS
from rdflib.namespace import SKOS, DC, DCTERMS
from typing import List


__all__ = [
    "Collection",
    "ResourceCollection"
]


_ns_hydra_str = str(RDF_NAMESPACES.HYDRA)
_ns_cts_str = str(RDF_NAMESPACES.CTS)
_ns_dts_str = str(RDF_NAMESPACES.DTS)
_ns_dct_str = str(DCTERMS)
_ns_cap_str = str(RDF_NAMESPACES.CAPITAINS)
_ns_rdf_str = str(RDF)
_ns_rdfs_str = str(RDFS)


class Collection(Exportable):
    """ Collection represents any resource's metadata. It has members and parents

    :ivar properties: Properties of the collection
    :type properties: dict
    :ivar parents: Parent of the node from the direct parent to the highest ascendant
    :type parents: [Collection]
    :ivar metadata: Metadata
    :type metadata: Metadata
    """
    TYPE_URI = URIRef(RDF_NAMESPACES.DTS.collection)
    MODEL_URI = URIRef(RDF_NAMESPACES.DTS.collection)
    EXPORT_TO = [Mimetypes.JSON.LD, Mimetypes.JSON.DTS.Std, Mimetypes.XML.RDF]

    def __init__(self, identifier="", *args, **kwargs):
        super(Collection, self).__init__(identifier, *args, **kwargs)
        self._graph = get_graph()

        self._node = URIRef(identifier)
        self._metadata = Metadata(node=self.asNode())

        self.graph.set((self.asNode(), RDF.type, self.TYPE_URI))
        self.graph.set((self.asNode(), RDF_NAMESPACES.DTS.model, self.MODEL_URI))

        self._parent = None
        self._children = {}

    def __repr__(self):
        return "%s(%s)#%s" % (self.__class__.__name__, self.id, id(self))

    @property
    def version(self):
        for x in self.graph.objects(self.asNode(), RDF_NAMESPACES.DTS.version):
            return x

    @version.setter
    def version(self, value):
        if not isinstance(value, Literal):
            value = Literal(value)
        self.graph.set((self.asNode(), RDF_NAMESPACES.DTS.version, value))

    @property
    def type(self):
        return list(self.graph.objects(self.asNode(), RDF.type))[0]

    @type.setter
    def type(self, value):
        if not isinstance(value, URIRef):
            value = URIRef(value)
        self.graph.set((self.asNode(), RDF.type, value))

    @property
    def model(self):
        return list(self.graph.objects(self.asNode(), RDF_NAMESPACES.DTS.model))[0]

    @model.setter
    def model(self, value):
        if not isinstance(value, URIRef):
            value = URIRef(value)
        self.graph.set((self.asNode(), RDF_NAMESPACES.DTS.model, value))

    @property
    def size(self):
        return len(self.members)

    # Graph Related Properties
    @property
    def graph(self):
        """ RDFLib Graph space

        :rtype: Graph
        """
        return self._graph

    @property
    def metadata(self):
        return self._metadata

    def asNode(self):
        """ Node representation of the collection in the graph

        :rtype: URIRef
        """
        return self._node

    @property
    def id(self):
        return str(self.asNode())

    def get_label(self, lang=None):
        """ Return label for given lang or any default

        :param lang: Language to request
        :return: Label value
        :rtype: Literal
        """
        x = None
        if lang is None:
            for obj in self.graph.objects(self.asNode(), RDFS.label):
                return obj
        for obj in self.graph.objects(self.asNode(), RDFS.label):
            x = obj
            if x.language == lang:
                return x
        return x

    def set_label(self, label, lang):
        """ Add the label of the collection in given lang

        :param label: Label Value
        :param lang:  Language code
        """
        self.metadata.add(SKOS.prefLabel, Literal(label, lang=lang))
        self.graph.addN([
            (self.asNode(), RDFS.label, Literal(label, lang=lang), self.graph),
        ])

    @property
    def children(self) -> dict:
        """ Dictionary of childrens {Identifier: Collection}

        :rtype: dict
        """
        return self._children

    @property
    def parents(self) -> List["Collection"]:
        """ Iterator to find parents of current collection, from closest to furthest

        :rtype: Generator[:class:`Collection`]
        """
        p = self.parent
        parents = []
        while p is not None:
            parents.append(p)
            p = p.parent
        return parents

    @property
    def parent(self):
        """ Parent of current object

        :rtype: Collection
        """
        return self._parent

    @parent.setter
    def parent(self, parent):
        """ Parents

        :param parent: Parent to set for the object
        :type parent: Collection
        :return:
        """
        self._parent = parent
        self.graph.add(
            (self.asNode(), RDF_NAMESPACES.CAPITAINS.parent, parent.asNode())
        )
        parent._add_member(self)

    def _add_member(self, member):
        """ Does not add member if it already knows it.

        .. warning:: It should not be called !

        :param member: Collection to add to members
        """
        if member.id in self.children:
            return None
        else:
            self.children[member.id] = member
            #self.graph.add((self.asNode(), RDF_NAMESPACES.DTS.child, member.asNode()))

    def __getitem__(self, key):
        """ Retrieve an item by its ID in the tree of a collection

        :param key: Key of the object to delete
        :return: Collection identified by the item
        """
        if key == self.id:
            return self
        for obj in self.members:
            if key == obj.id:
                return obj
        for obj in self.descendants:
            if obj.id == key:
                return obj
        raise UnknownCollection("%s is not part of this object" % key)

    def __delitem__(self, key):
        """ Delete a children of the collection

        :param key: Key of the object to delete
        """
        item = self[key]
        # Delete the graph Item
        self.graph.remove((item.asNode(), None, None))
        self.graph.remove((None, None, item.asNode()))
        self.metadata.remove()
        self.metadata.unlink()
        # Delete the Python item
        if len(item.parents) > 0:
            del item.parents[0].children[item.id]

    def __contains__(self, item):
        """ Retrieve an item by its ID in the tree of a collection

        :param item:
        :return: Collection identified by the item
        """
        for obj in self.descendants + [self]:
            if obj.id == item:
                return True
        return False

    @property
    def readable(self):
        """ Readable property should return elements where the element can be queried for getPassage / getReffs
        """
        return False

    @property
    def members(self):
        """ Children of the collection's item

        :rtype: [Collection]
        """
        return list(self.children.values())

    @property
    def descendants(self):
        """ Any descendant (no max level) of the collection's item

        :rtype: [Collection]
        """
        return self.members + \
            [submember for member in self.members for submember in member.descendants]

    @property
    def readableDescendants(self):
        """ List of element available which are readable

        :rtype: [Collection]
        """
        return [member for member in self.descendants if member.readable]

    def __namespaces_header__(self, cpt=None):
        """ Generates Namespaces Header given the graph

        :return: Dictionary with XMLNS prefix and uri as key and values
        """
        nm = self.graph.namespace_manager
        bindings = {}
        for predicate in set(self.graph.predicates()):
            prefix, namespace, name = nm.compute_qname(predicate)
            if prefix != "":
                bindings["xmlns:" + prefix] = str(URIRef(namespace))#[:-1]
            else:
                bindings["xmlns"] = str(URIRef(namespace))#[:-1]
        if cpt is True:
            bindings["xmlns:cpt"] = str(RDF_NAMESPACES.CAPITAINS)

        # Small hard coded fix for namespace that were not thought for RDF
        for k, v in bindings.items():
            if v in ["http://chs.harvard.edu/xmlns/cts/"]:
                bindings[k] = v[:-1]

        return bindings

    @classmethod
    def export_base_dts(cls, graph, obj, nsm):
        """ Export the base DTS information in a simple reusable way

        :param graph: Current graph where the information lie
        :param obj: Object for which we build info
        :param nsm: Namespace manager
        :return: Dict
        """

        o = {
            "@id": str(obj.asNode()),
            "@type": nsm.qname(obj.type),
            nsm.qname(RDF_NAMESPACES.HYDRA.title): str(obj.get_label()),
            nsm.qname(RDF_NAMESPACES.HYDRA.totalItems): obj.size
        }

        for desc in graph.objects(obj.asNode(), RDF_NAMESPACES.HYDRA.description):
            o[nsm.qname(RDF_NAMESPACES.HYDRA.description)] = str(desc)

        return o

    def __export__(self, output=None, namespace_manager=None):
        """ Export the collection item in the Mimetype required.

        ..note:: If current implementation does not have special mimetypes, reuses default_export method

        :param output: Mimetype to export to (Uses MyCapytain.common.utils.Mimetypes)
        :type output: str
        :return: Object using a different representation
        """

        if output == Mimetypes.JSON.DTS.Std:

            # Set-up a derived Namespace Manager
            if not namespace_manager:
                nsm = {
                    prefix: ns
                    for prefix, ns in self.graph.namespace_manager.namespaces()
                    if str(ns) not in [_ns_cap_str, _ns_cts_str, _ns_dts_str, _ns_dct_str, _ns_hydra_str]
                }
                nsm[""] = RDF_NAMESPACES.HYDRA
                nsm["cts"] = RDF_NAMESPACES.CTS
                nsm["dts"] = RDF_NAMESPACES.DTS
                nsm["dct"] = DCTERMS

            else:
                nsm = namespace_manager.namespaces()

            # Set-up a derived graph
            store = Subgraph(nsm)
            store.graphiter(self.graph, self.asNode(), ascendants=0, descendants=1)
            graph = store.graph
            nsm = store.graph.namespace_manager

            # Build the JSON-LD @context

            ignore_ns_for_bindings = [_ns_cap_str, _ns_hydra_str, _ns_rdf_str, _ns_rdfs_str]
            bindings = {}
            for predicate in set(graph.predicates()):
                prefix, namespace, name = nsm.compute_qname(predicate)
                if prefix not in bindings and str(namespace) not in ignore_ns_for_bindings:
                    bindings[prefix] = str(URIRef(namespace))

            # Builds the specific Store data
            extensions = {}
            dublincore = {}
            ignore_ns = [_ns_cap_str, _ns_hydra_str, _ns_rdf_str, _ns_rdfs_str, _ns_dts_str]

            # Builds the .dublincore and .extensions graphs
            for _, predicate, obj in store.graph:
                k = graph.qname(predicate)
                prefix, namespace, name = nsm.compute_qname(predicate)
                namespace = str(namespace)

                # Ignore namespaces that are part of the root DTS object
                if namespace in ignore_ns:
                    continue

                # Switch to the correct container depending on namespaces
                if namespace == str(DCTERMS):
                    metadata = dublincore
                else:
                    metadata = extensions

                if k in metadata:
                    if isinstance(metadata[k], list):
                        metadata[k].append(literal_to_dict(obj))
                    else:
                        metadata[k] = [metadata[k], literal_to_dict(obj)]
                else:
                    metadata[k] = literal_to_dict(obj)
                    if isinstance(metadata[k], dict):
                        metadata[k] = [metadata[k]]

            o = {"@context": bindings}
            o.update(self.export_base_dts(graph, self, nsm))
            o["@context"]["@vocab"] = _ns_hydra_str

            if extensions:
                o[graph.qname(RDF_NAMESPACES.DTS.extensions)] = extensions

            if dublincore:
                o[graph.qname(RDF_NAMESPACES.DTS.dublincore)] = dublincore

            if self.size:
                o[graph.qname(RDF_NAMESPACES.HYDRA.member)] = [
                    self.export_base_dts(self.graph, member, nsm)
                    for member in self.members
                ]

            # If the system handles citation structure
            if hasattr(self, "citation") and \
                    isinstance(self.citation, BaseCitationSet):
                if self.citation.depth:
                    o[graph.qname(RDF_NAMESPACES.DTS.term("citeDepth"))] = self.citation.depth

                if not self.citation.is_empty():
                    o[graph.qname(RDF_NAMESPACES.DTS.term("citeStructure"))] = self.citation.export(
                        Mimetypes.JSON.DTS.Std,
                        context=False,
                        namespace_manager=nsm
                    )

            del store
            return o
        elif output == Mimetypes.JSON.LD\
                or output == Mimetypes.XML.RDF:

            # We create a temp graph
            store = Subgraph(get_graph().namespace_manager)
            store.graphiter(self.graph, self.asNode(), ascendants=1, descendants=-1)

            o = store.serialize(format=RDFLIB_MAPPING[output], auto_compact=True, indent="")
            del store
            return o


class ResourceCollection(Collection):
    @property
    def lang(self):
        """ Languages this text is in

        :return: List of available languages
        """
        return str(self.graph.value(self.asNode(), DC.language))

    @lang.setter
    def lang(self, lang):
        """ Language this text is available in

        :param lang: Language to add
        :type lang: str
        """
        self.graph.set((self.asNode(), DC.language, Literal(lang)))

    def get_creator(self, lang=None):
        """ Get the DC Creator literal value

        :param lang: Language to retrieve
        :return: Creator string representation
        :rtype: Literal
        """
        return self.metadata.get_single(key=DC.creator, lang=lang)

    def get_title(self, lang=None):
        """ Get the title of the object

        :param lang: Lang to retrieve
        :return: Title string representation
        :rtype: Literal
        """
        return self.metadata.get_single(key=DC.title, lang=lang)

    def get_description(self, lang=None):
        """ Get the description of the object

        :param lang: Lang to retrieve
        :return: Description string representation
        :rtype: Literal
        """
        return self.metadata.get_single(key=DC.description, lang=lang)

    def get_subject(self, lang=None):
        """ Get the subject of the object

        :param lang: Lang to retrieve
        :return: Subject string representation
        :rtype: Literal
        """
        return self.metadata.get_single(key=DC.subject, lang=lang)
